#include <stdio.h>
#include <stdlib.h>

typedef struct
{
	short wallptr, wallnum;
	long ceilingz, floorz;
	short ceilingstat, floorstat;
	short ceilingpicnum, ceilingheinum;
	signed char ceilingshade;
	char ceilingpal, ceilingxpanning, ceilingypanning;
	short floorpicnum, floorheinum;
	signed char floorshade;
	char floorpal, floorxpanning, floorypanning;
	char visibility, filler;
	short lotag, hitag, extra;
} sectortype;



typedef struct
{
	long x, y;
	short point2, nextwall, nextsector, cstat;
	short picnum, overpicnum;
	signed char shade;
	char pal, xrepeat, yrepeat, xpanning, ypanning;
	short lotag, hitag, extra;
} walltype;

typedef struct
{
    long x, y, z;
	short cstat, picnum;
	signed char shade;
	char pal, clipdist, filler;
	unsigned char xrepeat, yrepeat;
	signed char xoffset, yoffset;
	short sectnum, statnum;
	short ang, owner, xvel, yvel, zvel;
	short lotag, hitag, extra;
} spritetype;

typedef struct
{
    int ver;
    int s_posx, s_posy, s_posz;
    short s_ang, s_sec;
	short secnum, walnum, sprnum;
	sectortype *sec;
	walltype *wal;
	spritetype *spr;
}maptype;

typedef struct
{
    char notmd2;
    short sprnum;
    short angoff;
}maphacktype;

int loadmap(char *fname, maptype *m){
    FILE *fp;

    // Open
    fp = fopen( fname, "rb");
    if(fp == NULL) return -1;
    
    // Load Version
    size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream);
    fread(&((*m).ver), sizeof(int), 1, fp);
    printf("version = %i\n", (*m).ver);
    
    // Load Start Position
    fread(&((*m).s_posx), sizeof(int), 1, fp);
    printf("Player starts at (%i,", (*m).s_posx);
    fread(&((*m).s_posy), sizeof(int), 1, fp);
    printf(" %i,", (*m).s_posy);
    fread(&((*m).s_posz), sizeof(int), 1, fp); //Note: Z coordinates are all shifted up 4
    printf(" %i)", (*m).s_posz);  
    fread(&((*m).s_ang), sizeof(char), 2, fp); //All angles are from 0-2047, clockwise
    printf(" facing %i", (*m).s_ang);
	fread(&((*m).s_sec), sizeof(char), 2, fp); //Sector of starting point
	printf(" in sector %i.\n", (*m).s_sec);
    
    // Load Sectors
    fread(&((*m).secnum), sizeof(char), 2, fp);
    printf("Map has %i sectors,", (*m).secnum);
    (*m).sec = malloc(sizeof(sectortype)*(*m).secnum);
	fread((*m).sec, sizeof(sectortype), (*m).secnum, fp);
	
    // Load Walls
    fread(&((*m).walnum), sizeof(char), 2, fp);
    printf(" %i walls", (*m).walnum);
    (*m).wal = malloc(sizeof(walltype)*(*m).walnum);
	fread((*m).wal, sizeof(walltype), (*m).walnum, fp);
    
    // Load Sprites
    fread(&((*m).sprnum), sizeof(char), 2, fp);
    printf(" and %i sprites.\n", (*m).sprnum);
    (*m).spr = malloc(sizeof(spritetype)*(*m).sprnum);
	fread((*m).spr, sizeof(spritetype), (*m).sprnum, fp);
	
    
    // Close
    fclose(fp);
    return 0;
}

int parseline(char *line, maphacktype *mh){
    char sprite_l[] = "sprite";
    char sprite_u[] = "SPRITE";
    char angoff_l[] = "angoff";
    char angoff_u[] = "ANGOFF";
    char notmd2_l[] = "notmd2";
    char notmd2_u[] = "NOTMD2";
    char intstr[5];
    
    int i, j, k;
    short tmp, sign = 1;
    
    // check first key word, should be sprite else return null
    
    // Skip spaces
    i = 0;
    while(line[i] == ' '){
        i++;
    }
    //printf("Skipped %i spaces\n", i);
    //check keyword
    j = 0;
    while(line[i] != ' '){
        //printf("'%c'", line[i]);
        if(line[i] != sprite_l[j] && line[i] != sprite_u[j]){
            if(line[i] == '/'){
                //printf("Comment\n");
                return 1;
            }
            else if(line[i] == '\n'){
                //printf("Empty line\n");
                return 2;
            }
            else{
                printf("Bad keyword!\n");
                return -1;
            }
        }
        i++;
        j++;
    }
    //printf("Keyword sprite found\n");
    // parse the sprite number, if no sprite number  return null
    
    // Skip spaces
    while(line[i] == ' '){
        i++;
    }
    //printf("Skipped %i spaces\n", i);
    // parse sprite number
    j =  0;
    while(line[i] != ' '){
        if(line[i] < '0' || line[i] > '9' ){
            printf("Expected int but found '%c'\n", line[i]);
            return -1;
        }
        //printf("Found int : %c\n", line[i]);
        intstr[j] = line[i];
        i++;
        j++;
    }
    intstr[j] = '\0';
    //printf("Sprite number parsed : %s\n", intstr);
    // convert string to int
    k = 1;
    tmp = 0;
    j --;
    while(j >= 0){
        tmp = tmp + ((intstr[j] - '0') * k);
        //printf("tmp = %i, last number = %c\n", tmp, intstr[j]);
        j--;
        k = k * 10;
    }
    //printf("Sprite number converted to int : %i\n", tmp);
    mh->sprnum = tmp;
    // check hack command if not notmd2 or angoff return null
    
    while(line[i] == ' '){
        i++;
    }
    //printf("Skipped %i spaces\n", i);
    
    // angoff
    if(line[i] == 'a' || line[i] == 'A'){
        j = 0;
        while(line[i] != ' '){
            if(line[i] != angoff_l[j] && line[i] != angoff_u[j]){
                printf("Failed to parse angoff command!\n");
                return -1;
            }
            i++;
            j++;
        }
        //printf("angoff command found\n");
        // if angoff parse offset, if no offsett return null
        // parse sprite number
        
        while(line[i] == ' '){
            i++;
        }
        //printf("Skipped %i spaces\n", i);
        
        //identify sign
        
        if(line[i] == '+'){
            sign = 1;
            i++;
        }
        else if(line[i] == '-'){
            sign = -1;
            i++;
        }
        else{
            sign = 1;
        }
        
        j =  0;
        while(line[i] != ' ' && line[i] != '\n' && line[i] != '\0'){
            if(line[i] < '0' || line[i] > '9' ){
                printf("angoff parseing failed (%c is not a number)\n", line[i]);
                return -1;
            }
            intstr[j] = line[i];
            i++;
            j++;
        }
        // convert string to int
        k = 1;
        tmp = 0;
        j--;
        while(j >= 0){
            tmp = tmp + ((intstr[j] - '0') * k);
            j--;
            k = k * 10;
        }
        tmp = tmp * sign;
        //printf("angoff parsed as : %i\n", tmp);
        mh->angoff = tmp;
        mh->notmd2 = 0;
    }
    // notmd2
    else if(line[i] == 'n' || line[i] == 'N'){
        j = 0;
        while(line[i] != ' ' && line[i] != '\n' && line[i] != '\0'){
            if(line[i] != notmd2_l[j] && line[i] != notmd2_u[j]){
                printf("Failed to parse notmd2 command (%c did not match %c)\n", line[i], notmd2_l[j]);
                return -1;
            }
            i++;
            j++;
        }
        mh->angoff = 0;
        mh->notmd2 = 1;
    }
    
    return 0;
}

int loadmaphack(char *fname, maphacktype *mh[]){
    FILE *fp;
    maphacktype mptmp;
    char buf[256];
    int i = 0, r;
    
    mh[0] = NULL;
    
    // Open
    fp = fopen( fname, "r");
    if(fp == NULL) return -1;
    while(fgets(buf, 256, fp) != NULL){
        //printf("Line %i\n", i);
        if((r = parseline(buf, &mptmp)) == -1){
            printf("parse failed at line %i!\n", i);
            return -1;
        }
        else if(r == 0){
            mh[i] = malloc(sizeof(maphacktype));
            *(mh[i]) = mptmp;
            i++;
        }
    }
    mh[i] = NULL;
    // Close
    fclose(fp);
    return 0;
}

int main(int argc, char *argv[]){
    maptype hackedmap, nonhackedmap;
    maphacktype *mh[1024];
    int i = 0, j = 0;
    char oldhack[512], newhack[512];
    FILE *fp;
    printf("\n== Hacked map ==\n");
    if(loadmap(argv[1], &hackedmap) == -1) return -1;
    printf("\n== To be Hacked map ==\n");
    if(loadmap(argv[2], &nonhackedmap) == -1) return -1;
    printf("\n\n");
    strcpy(oldhack, argv[1]);
    strcpy(newhack, argv[2]);
    oldhack[strlen(argv[1])-1] = 'k';
    oldhack[strlen(argv[1])-2] = 'h';
    oldhack[strlen(argv[1])-3] = 'm'; 
    newhack[strlen(argv[2])-1] = 'k';
    newhack[strlen(argv[2])-2] = 'h';
    newhack[strlen(argv[2])-3] = 'm';
    if(loadmaphack(oldhack, mh) == -1) return -1;
    
    // Open
    fp = fopen( newhack, "w");
    if(fp == NULL) return -1;
    
    i = 0;
    while(mh[i] != NULL){  
        if(mh[i]->sprnum >= hackedmap.sprnum){
            printf("sprite %i ", mh[i]->sprnum);
            if(mh[i]->notmd2){
                printf("notmd2 does not exist\n");
            }
            else{
                printf("angoff %i does not exist\n", mh[i]->angoff);
            }
            i++;
            continue;
        }
        j = 0;
        //printf("=== maphack %i ===\n", i);
        
        //printf("map has %i sprites\n", nonhackedmap.sprnum);
        while(j < nonhackedmap.sprnum){
            //printf("sprite %i with tile %i\n", j, nonhackedmap.spr[j].picnum);
            if(nonhackedmap.spr[j].x == hackedmap.spr[mh[i]->sprnum].x && 
                nonhackedmap.spr[j].y == hackedmap.spr[mh[i]->sprnum].y &&
                nonhackedmap.spr[j].z == hackedmap.spr[mh[i]->sprnum].z &&
                nonhackedmap.spr[j].picnum == hackedmap.spr[mh[i]->sprnum].picnum &&
                nonhackedmap.spr[j].ang == hackedmap.spr[mh[i]->sprnum].ang){
                    // Found a match
                    printf("converted ");
                    printf("sprite %i ", mh[i]->sprnum);
                    if(mh[i]->notmd2){
                        printf("notmd2 ");
                    }
                    else{
                        printf("angoff %i ", mh[i]->angoff);
                    }
                    printf("Succesfully\n");
                    fprintf(fp, "sprite %i ", j);
                    if(mh[i]->notmd2){
                        fprintf(fp, "notmd2\n");
                    }
                    else{
                        fprintf(fp, "angoff %i\n", hackedmap.spr[mh[i]->sprnum].ang + mh[i]->angoff - nonhackedmap.spr[j].ang);
                    }
                    break;
                }
            j++;
        }
        if(j == nonhackedmap.sprnum){
            printf("Failed to find match for ");
            printf("sprite %i ", mh[i]->sprnum);
                if(mh[i]->notmd2){
                    printf("notmd2\n");
                }
                else{
                    printf("angoff %i\n", mh[i]->angoff);
                }
        }
        i++;
    }
    
    fclose(fp);
    return 0;
}
